# coding: utf-8

"""
Работа с подполями.
"""

from typing import TYPE_CHECKING
from irbis.abstract import Hashable
from irbis.records.abstract import ValueMixin
if TYPE_CHECKING:
    from typing import List, Optional

    Value = Optional[str]
    SubFieldList = List['SubField']


class SubField(Hashable, ValueMixin):
    """
    MARC record subfield with code and text value.
    """

    DEFAULT_CODE = '\0'

    __slots__ = 'code', 'value'

    def __init__(self, code: str = DEFAULT_CODE,
                 value: 'Value' = None) -> None:
        self.code: str = self.validate_code(code) or SubField.DEFAULT_CODE
        self.value: 'Value' = self.validate_value(value)

    def assign_from(self, other: 'SubField') -> None:
        """
        Присваивание от другого поля: код и значение
        берутся от другого подполя.

        :param other: Другое подполе
        :return: None
        """
        self.code = other.code
        self.value = other.value

    def clone(self) -> 'SubField':
        """
        Клонирование подполя.

        :return: Клон подполя
        """
        return SubField(self.code, self.value)

    @property
    def data(self):
        """
        Динамическое свойство извлечения данных в представлении стандартных
        типов данных Python.

        :return: строка со значением подполя
        """
        return self.value

    def __str__(self):
        if self.code == self.DEFAULT_CODE:
            return ''
        return '^' + self.code + (self.value or '')

    def __bool__(self):
        return self.code != self.DEFAULT_CODE and bool(self.value)

    def __hash__(self):
        return hash((self.code, self.value))

    @staticmethod
    def validate_code(code: str) -> str:
        """
        Валидация кода подполя на тип (строка) и длинну (1 символ)

        :code: код подполя
        :return: код подполя
        """
        if isinstance(code, str) is False:
            raise ValueError('Код подполя должен быть строкой')
        if len(code) != 1:
            raise ValueError('Код подполя должен быть односимвольным')
        return code.lower()



# coding: utf-8

"""
Модуль реализующий абстрактные классы пакета irbis.records
"""

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING
from irbis._common import LOGICALLY_DELETED, PHYSICALLY_DELETED
if TYPE_CHECKING:
    from typing import Any, List, Optional


class ValueMixin:
    """
    Примесь для проверки атрибута value
    """
    @staticmethod
    def validate_value(value: 'Any') -> 'Optional[str]':
        """
        Валидация Field.value и SubField.value

        :value: значение
        """
        if value is None:
            return value
        if isinstance(value, str):
            if value:
                return value
            raise ValueError('value не может быть пустой строкой')
        raise TypeError('Не поддерживаемый тип value')


class AbstractRecord:
    """
    Абстрактный класс с общими свойствами и методами для классов Record
    и RawRecord.
    """
    __metaclass__ = ABCMeta
    field_type: 'Any'

    @abstractmethod
    def __init__(self, *args: 'Any'):
        self.database: 'Optional[str]' = None
        self.mfn = 0
        self.version = 0
        self.status = 0
        self.fields: 'Any' = []
        self.__bulk_set__(*args)

    def clear(self) -> 'AbstractRecord':
        """
        Очистка записи (удаление всех полей).

        :return: Self
        """
        self.fields.clear()
        return self

    def clone(self) -> 'AbstractRecord':
        """
        Клонирование записи.

        :return: Полный клон записи
        """
        result = self.__class__()
        result.database = self.database
        result.mfn = self.mfn
        result.status = self.status
        result.version = self.version
        result.fields = self.clone_fields()
        return result

    @abstractmethod
    def clone_fields(self) -> 'List[Any]':
        """
        Абстрактный метод клонирования полей записи
        :return: список полей
        """

    def encode(self) -> 'List[str]':
        """
        Кодирование записи в серверное представление.

        :return: Список строк
        """
        result = [str(self.mfn) + '#' + str(self.status),
                  '0#' + str(self.version)]
        for field in self.fields:
            result.append(str(field))
        return result

    def is_deleted(self) -> bool:
        """
        Удалена ли запись?
        :return: True для удаленной записи
        """
        return (self.status & (LOGICALLY_DELETED | PHYSICALLY_DELETED)) != 0

    def parse(self, text: 'List[str]') -> None:
        """
        Разбор текстового представления записи (в серверном формате).

        :param text: Список строк
        :return: None
        """
        if text:
            line = text[0]
            parts = line.split('#')
            self.mfn = int(parts[0])
            if len(parts) != 1 and parts[1]:
                self.status = int(parts[1])
            line = text[1]
            parts = line.split('#')
            self.version = int(parts[1])
            self.fields.clear()
            for line in text[2:]:
                if line:
                    self.parse_line(line)
        else:
            raise ValueError('text argument is empty')

    @abstractmethod
    def parse_line(self, line: str) -> None:
        """
        Абстрактный метод разбора строки из текстового представления записи
        (в серверном формате). Реализуется у дочерних классов Record
        и RawRecord.

        :param line: строка
        :return: None
        """

    def remove_at(self, index: int) -> 'AbstractRecord':
        """
        Удаление поля в указанной позиции.

        :param index: Позиция для удаления.
        :return: self
        """
        assert 0 <= index < len(self.fields)

        self.fields.remove(self.fields[index])
        return self

    def reset(self) -> 'AbstractRecord':
        """
        Сбрасывает состояние записи, отвязывая её от базы данных.
        Поля при этом остаются нетронутыми.
        :return: Self.
        """
        self.mfn = 0
        self.status = 0
        self.version = 0
        self.database = None
        return self

    @abstractmethod
    def __bulk_set__(self, *args):
        """
        Абстрактный приватный метод полей записи. Реализуется у дочерних
        классов.

        Внимание. Пользователь не должен явно обращаться к данному методу.

        :param args: Аргументы для [пере]создания полей
        :return: ничего
        """

    def __bool__(self):
        return bool(self.fields)

    def __len__(self):
        return len(self.fields)

    def __str__(self):
        return '\n'.join(str(field) for field in self.fields)



# coding: utf-8

"""
Работа с терминами словаря.
"""

from typing import TYPE_CHECKING
from irbis._common import safe_str
if TYPE_CHECKING:
    from typing import Iterable, List, Optional


class PostingParameters:
    """
    Параметры для команды ReadPostings.
    """

    __slots__ = 'database', 'first', 'fmt', 'number', 'terms'

    def __init__(self, term: 'Optional[str]' = None,
                 fmt: 'Optional[str]' = None) -> None:
        self.database: 'Optional[str]' = None
        self.first: int = 1
        self.fmt: 'Optional[str]' = fmt
        self.number: int = 0
        self.terms: 'List[str]' = []
        if term:
            self.terms.append(term)

    def __str__(self):
        return str(self.terms)


class TermInfo:
    """
    Информация о поисковом терме.
    """

    __slots__ = 'count', 'text'

    def __init__(self, count: int = 0, text: str = '') -> None:
        self.count: int = count
        self.text: str = text

    @staticmethod
    def parse(lines: 'Iterable[str]'):
        """
        Parse the text for term info.

        :param lines: Text to parse
        :return: Term info
        """
        result = []
        for line in lines:
            parts = line.split('#', 1)
            item = TermInfo(int(parts[0]), parts[1])
            result.append(item)
        return result

    def __str__(self):
        return str(self.count) + '#' + self.text


class TermParameters:
    """
    Параметры для команды ReadTerms
    """

    __slots__ = 'database', 'number', 'reverse', 'start', 'format'

    def __init__(self, start: 'Optional[str]' = None,
                 number: int = 10) -> None:
        self.database: str = ''
        self.number: int = number
        self.reverse: bool = False
        self.start: 'Optional[str]' = start
        self.format: 'Optional[str]' = None

    def __str__(self):
        return str(self.number) + ' ' + safe_str(self.format)


class TermPosting:
    """
    Постинг терма.
    """

    __slots__ = 'mfn', 'tag', 'occurrence', 'count', 'text'

    def __init__(self) -> None:
        self.mfn: int = 0
        self.tag: int = 0
        self.occurrence: int = 0
        self.count: int = 0
        self.text: 'Optional[str]' = None

    def parse(self, text: str) -> None:
        """
        Parse the text for term postings.

        :param text: Text to parse
        :return: None
        """

        parts = text.split('#', 4)
        if len(parts) >= 4:
            self.mfn = int(parts[0])
            self.tag = int(parts[1])
            self.occurrence = int(parts[2])
            self.count = int(parts[3])
            if len(parts) > 4:
                self.text = parts[4]
        else:
            # Ууточнить, требуется ли бросать исключение
            pass

    def __str__(self):
        subst = ''
        if self.text:
            subst = self.text
        return ' '.join([str(self.mfn), str(self.tag),
                         str(self.occurrence), str(self.count),
                         subst])


__all__ = ['PostingParameters', 'TermInfo', 'TermParameters', 'TermPosting']



# coding: utf-8

"""
Ресурсы сервера.
"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Dict, List, Optional


class Resource:
    """
    Текстовый файл на сервере.
    """

    __slots__ = 'name', 'content'

    def __init__(self, name: str, content: str) -> None:
        self.name: str = name
        self.content: str = content

    def __str__(self) -> str:
        return f"{self.name}: {self.content}"


class ResourceDictionary:
    """
    Словарь текстовых ресурсов.
    """

    __slots__ = ('dictionary',)

    def __init__(self) -> None:
        self.dictionary: 'Dict[str, Resource]' = {}

    def add(self, name: str, content: str) -> 'ResourceDictionary':
        """
        Регистрация ресурса в словаре.
        :param name: Имя
        :param content: Содержимое
        :return: Словарь
        """
        self.dictionary[name] = Resource(name, content)
        return self

    def all(self) -> 'List[Resource]':
        """
        Все зарегистрированные ресурсы в виде массива.
        :return: Массив
        """
        result: 'List[Resource]' = []
        for item in self.dictionary.values():
            result.append(item)
        return result

    def clear(self) -> 'ResourceDictionary':
        """
        Очистка словаря.
        :return: Словарь
        """
        self.dictionary.clear()
        return self

    def count(self) -> int:
        """
        Вычисление длины словаря.
        :return: Число хранимых ресурсов
        """
        return len(self.dictionary)

    def get(self, name: str) -> 'Optional[str]':
        """
        Получение ресурса из словаря по имени.
        :param name: Имя
        :return: Содержимое ресурса либо None
        """
        if name in self.dictionary:
            return self.dictionary[name].content
        return None

    def have(self, name: str) -> bool:
        """
        Есть ли элемент с указанным именем в словаре?
        :param name: Имя
        :return: Наличие в словаре
        """
        return name in self.dictionary

    def put(self, name: str, content: str) -> 'ResourceDictionary':
        """
        Добавление ресурса в словарь.
        :param name: Имя
        :param content: Содержимое
        :return: Словарь
        """
        self.dictionary[name] = Resource(name, content)
        return self

    def remove(self, name: str) -> 'ResourceDictionary':
        """
        Удаление ресурса из словаря.
        :param name: Имя
        :return: Словарь
        """
        del self.dictionary[name]
        return self


__all__ = ['Resource', 'ResourceDictionary']



# coding: utf-8

"""
Класс, умеющий сохранять свое содержимое
в текстовом файле.
"""

from irbis._common import ANSI


class Saveable:
    """
    A class that can store its content in a text file.
    """

    def save(self, filename: str) -> None:
        """
        Save the instance to the specified file.

        :param filename: Name of the file
        :return: None
        """

        with open(filename, 'wt', encoding=ANSI) as stream:
            text = str(self)
            stream.write(text)



# coding: utf-8

"""
Работа с PAR-файлами.
"""

from collections import defaultdict
from typing import TYPE_CHECKING
from irbis._common import ANSI
if TYPE_CHECKING:
    from typing import DefaultDict, Iterable


class ParFile:
    """
    PAR-файл.
    """

    __slots__ = ('xrf', 'mst', 'cnt', 'n01', 'n02', 'l01', 'l02', 'ifp',
                 'any', 'pft', 'ext')

    def __init__(self, mst: str = '') -> None:
        self.xrf: str = mst
        self.mst: str = mst
        self.cnt: str = mst
        self.n01: str = mst
        self.n02: str = mst
        self.l01: str = mst
        self.l02: str = mst
        self.ifp: str = mst
        self.any: str = mst
        self.pft: str = mst
        self.ext: str = mst

    @staticmethod
    def make_dict(text: 'Iterable[str]') -> 'DefaultDict':
        """
        Make the dictionary from the text.

        :param text: Text to parse.
        :return: Dictionary
        """
        result: 'DefaultDict' = defaultdict(lambda: '')
        for line in text:
            if not line:
                continue
            parts = line.split('=', 1)
            if len(parts) < 2:
                continue
            key = parts[0].strip()
            value = parts[1].strip()
            result[key] = value
        return result

    def parse(self, text: 'Iterable[str]') -> None:
        """
        Parse the text for PAR entries.

        :param text: Text to parse
        :return: None
        """

        paths = ParFile.make_dict(text)
        self.xrf = paths['1']
        self.mst = paths['2']
        self.cnt = paths['3']
        self.n01 = paths['4']
        self.n02 = paths['5']
        self.l01 = paths['6']
        self.l02 = paths['7']
        self.ifp = paths['8']
        self.any = paths['9']
        self.pft = paths['10']
        self.ext = paths['11']

    def save(self, filename: str) -> None:
        """
        Save paths to the specified file.

        :param filename: File to use
        :return: None
        """
        with open(filename, 'wt', encoding=ANSI) as stream:
            stream.write(f'1={self.xrf}\n')
            stream.write(f'2={self.mst}\n')
            stream.write(f'3={self.cnt}\n')
            stream.write(f'4={self.n01}\n')
            stream.write(f'5={self.n02}\n')
            stream.write(f'6={self.l01}\n')
            stream.write(f'7={self.l02}\n')
            stream.write(f'8={self.ifp}\n')
            stream.write(f'9={self.any}\n')
            stream.write(f'10={self.pft}\n')
            stream.write(f'11={self.ext}\n')

    def __str__(self):
        result = ['1=' + self.xrf,
                  '2=' + self.mst,
                  '3=' + self.cnt,
                  '4=' + self.n01,
                  '5=' + self.n02,
                  '6=' + self.l01,
                  '7=' + self.l02,
                  '8=' + self.ifp,
                  '9=' + self.any,
                  '10=' + self.pft,
                  '11=' + self.ext]
        return '\n'.join(result)


def load_par_file(filename: str) -> ParFile:
    """
    Load PAR from the specified file.

    :param filename: File to use
    :return: PAR file
    """

    result = ParFile()
    with open(filename, 'rt', encoding=ANSI) as stream:
        lines = stream.readlines()
        result.parse(lines)
    return result


__all__ = ['load_par_file', 'ParFile']



# coding: utf-8

"""
Модуль, реализующий функционал для объектно-ориентированного представления
записей, полей и подполей из Ирбис64
"""

from irbis.records.abstract import AbstractRecord
from irbis.records.raw_record import RawRecord
from irbis.records.record import Record
from irbis.records.field import Field
from irbis.records.subfield import SubField


__all__ = ['AbstractRecord', 'RawRecord', 'Record', 'Field', 'SubField']



# coding: utf-8

"""
Работа с MNU-файлами.
"""

from typing import TYPE_CHECKING
from irbis._common import ANSI, STOP_MARKER
if TYPE_CHECKING:
    from typing import List, Optional


class MenuEntry:
    """
    Пара строк в меню.
    """

    __slots__ = 'code', 'comment'

    def __init__(self, code: str = '', comment: str = '') -> None:
        self.code: str = code
        self.comment: str = comment

    def __str__(self):
        if self.comment:
            return self.code + ' - ' + self.comment
        return self.code

    def __repr__(self):
        return self.__str__()


class MenuFile:
    """
    Файл меню.
    """

    __slots__ = ('entries',)

    def __init__(self) -> None:
        self.entries: 'List[MenuEntry]' = []

    def add(self, code: str, comment: str = ''):
        """
        Add an entry to the menu.

        :param code: Code
        :param comment: Comment
        :return: Self
        """
        entry = MenuEntry(code, comment)
        self.entries.append(entry)
        return self

    def get_entry(self, code: str) -> 'Optional[MenuEntry]':
        """
        Get an entry for the specified code.

        :param code: Code to search
        :return: Found entry or None
        """

        code = code.lower()
        for entry in self.entries:
            if entry.code.lower() == code:
                return entry

        code = code.strip()
        for entry in self.entries:
            if entry.code.lower() == code:
                return entry

        code = MenuFile.trim_code(code)
        for entry in self.entries:
            if entry.code.lower() == code:
                return entry

        return None

    def get_value(self, code: str,
                  default_value: 'Optional[str]' = None) -> 'Optional[str]':
        """
        Get value for the specified code.

        :param code: Code to search
        :param default_value: Default value
        :return: Found or default value
        """

        entry = self.get_entry(code)
        result = entry.comment if entry else default_value
        return result

    def parse(self, lines: 'List[str]') -> None:
        """
        Parse the text for menu entries.

        :param lines: Text to parse
        :return: None
        """

        i = 0
        while i + 1 < len(lines):
            code = lines[i]
            comment = lines[i + 1]
            if code.startswith(STOP_MARKER):
                break
            self.add(code, comment)
            i += 2

    def save(self, filename: str) -> None:
        """
        Save the menu to the specified file.

        :param filename: Name of the file
        :return: None
        """

        with open(filename, 'wt', encoding=ANSI) as stream:
            for entry in self.entries:
                stream.write(entry.code + '\n')
                stream.write(entry.comment + '\n')
            stream.write(STOP_MARKER + '\n')

    @staticmethod
    def trim_code(code: str) -> str:
        """
        Trim the code.

        :param code: code to process
        :return: Trimmed code
        """

        result = code.strip(' -=:')
        return result

    def __str__(self):
        result = []
        for entry in self.entries:
            result.append(entry.code)
            result.append(entry.comment)
        result.append(STOP_MARKER)
        return '\n'.join(result)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        yield from self.entries

    def __getitem__(self, item: str):
        return self.get_value(item)


def load_menu(filename: str) -> MenuFile:
    """
    Чтение меню из файла.

    :param filename: Имя файла
    :return: Меню
    """
    with open(filename, 'rt', encoding=ANSI) as stream:
        result = MenuFile()
        lines = stream.readlines()
        result.parse(lines)
        return result


__all__ = ['load_menu', 'MenuEntry', 'MenuFile']



# coding: utf-8

"""
Информация о базах данных.
"""

from typing import TYPE_CHECKING
from irbis._common import SHORT_DELIMITER
from irbis.response import ServerResponse
if TYPE_CHECKING:
    from typing import List, Optional


class DatabaseInfo:
    """
    Информация о базе данных ИРБИС.
    """

    __slots__ = ('name', 'description', 'max_mfn', 'logically_deleted',
                 'physically_deleted', 'nonactualized', 'locked_records',
                 'database_locked', 'read_only')

    def __init__(self, name: 'Optional[str]' = None,
                 description: 'Optional[str]' = None) -> None:
        self.name: 'Optional[str]' = name
        self.description: 'Optional[str]' = description
        self.max_mfn: int = 0
        self.logically_deleted: 'List[int]' = []
        self.physically_deleted: 'List[int]' = []
        self.nonactualized: 'List[int]' = []
        self.locked_records: 'List[int]' = []
        self.database_locked: bool = False
        self.read_only: bool = False

    @staticmethod
    def _parse(line: str) -> 'List[int]':
        if not line:
            return []
        return [int(x) for x in line.split(SHORT_DELIMITER) if x]

    def parse(self, response: ServerResponse) -> None:
        """
        Parse the server response for database info.

        :param response: Response to parse
        :return: None
        """
        self.logically_deleted = self._parse(response.ansi())
        self.physically_deleted = self._parse(response.ansi())
        self.nonactualized = self._parse(response.ansi())
        self.locked_records = self._parse(response.ansi())
        self.max_mfn = int(response.ansi())
        self.database_locked = bool(int(response.ansi()))

    def __str__(self):
        if not self.description:
            return self.name or '(none)'
        return self.name + ' - ' + self.description


__all__ = ['DatabaseInfo']



# coding: utf-8

"""
Reading and writing plain text files in IRBIS format.
"""

# pylint: disable=too-many-locals
# pylint: disable=too-many-statements

from typing import TYPE_CHECKING
from cbsvibpyirbis._common import ANSI, STOP_MARKER, safe_str
from cbsvibpyirbis.error import IrbisError
from cbsvibpyirbis.records import SubField, Field, Record
if TYPE_CHECKING:
    from typing import Iterable, Optional, List


def read_text_record(stream) -> 'Optional[Record]':
    """
    Чтение записи из файла в текстовом обменном формате ИРБИС.

    :param stream: Файл
    :return: Запись или None
    """

    result = Record()
    while True:
        line: str = stream.readline()
        if not line:
            break
        line = line.strip()
        if line.startswith(STOP_MARKER):
            break
        if not line.startswith('#'):
            break
        parts = line[1:].split(':', 1)
        if len(parts) != 2:
            break
        tag = int(parts[0])
        text = parts[1][1:]
        field = Field(tag)
        field.parse(text)
        result.fields.append(field)

    if not result.fields:  # Если в записи нет полей, возвращаем None
        return None

    return result


def write_text_record(stream, record: Record) -> None:
    """
    Сохранение записи в файл в текстовом обменном формате ИРБИС.

    :param stream: Текстовый поток, в который разрешена запись.
    :param record: Библиографическая запись.
    :return: None
    """

    assert stream
    assert record

    for field in record.fields:
        parts = ['#' + str(field.tag) + ': ' + safe_str(field.value)]
        for subfield in field.subfields:
            parts.extend(str(subfield))
        line = ''.join(parts) + '\n'
        stream.write(line)
    stream.write(STOP_MARKER + '\n')


###############################################################################

# Length of the record marker
MARKER_LENGTH = 24

# Record delimiter
RECORD_DELIMITER = 0x1D

# Field delimiter
FIELD_DELIMITER = 0x1E

# Subfield delimiter
SUBFIELD_DELIMITER = 0x1F


def parse_int(buffer: 'Iterable'):
    """
    Parse the bytes for integer value.

    :param buffer: Buffer to parse
    :return: Integer value
    """

    result = 0
    for byte in buffer:
        result = result * 10 + byte - 48
    return result


def encode_int(buffer: bytearray, position: int,
               length: int, value: int) -> None:
    """
    Encode the integer value.

    :param buffer: Buffer to fill
    :param position: Start position
    :param length: Chunk length
    :param value: Value to encode
    :return: None
    """

    length -= 1
    position += length
    while length >= 0:
        buffer[position] = value % 10 + ord('0')
        value //= 10
        position -= 1
        length -= 1


def encode_str(buffer: bytearray, position: int,
               value: 'Optional[str]', encoding: str) -> int:
    """
    Encode the string value.

    :param buffer: Buffer to fill
    :param position: Start position
    :param value: Value to encode (can be None or empty string)
    :param encoding: Encoding to use
    :return: Updated position
    """

    if value:
        encoded = value.encode(encoding)
        for byte in encoded:
            buffer[position] = byte
            position += 1

    return position


def read_iso_record(stream, charset: str = ANSI) -> 'Optional[Record]':
    """
    Чтение записи из файла в формате ISO 2709.

    :param stream: Файл или файлоподобный объект
    :param charset: Кодировка
    :return: Декодированная запись либо None
    """

    # Считываем длину записи
    marker = stream.read(5)
    if len(marker) != 5:
        return None

    # а затем и ее остаток
    record_length = parse_int(marker)
    need = record_length - 5
    tail = stream.read(need)
    if len(tail) != need:
        return None

    # Простая проверка, что мы имеем дело с нормальной ISO-записью
    record = marker + tail
    if record[record_length - 1] != RECORD_DELIMITER:
        return None

    # Превращаем запись в Unicode
    indicator_length = parse_int(record[10:11])
    base_address = parse_int(record[12:17])

    # Начинаем собственно конверсию
    result = Record()

    # Пошли по полям при помощи справочника
    directory = MARKER_LENGTH
    while record[directory] != FIELD_DELIMITER:
        # если нарвались на разделитель, значит, справочник закончился

        tag = parse_int(record[directory:directory + 3])
        field_length = parse_int(record[directory + 3:directory + 7])
        field_offset = parse_int(record[directory + 7:directory + 12]) + \
            base_address
        field = Field(tag)
        result.fields.append(field)

        if tag < 10:
            # фиксированное поле
            # не может содержать подполей и индикаторов
            field.value = record[field_offset:field_offset +
                                 field_length - 1].decode(charset)
        else:
            # поле переменной длины
            # содержит два однобайтных индикатора
            # может содержать подполя
            start = field_offset + indicator_length
            stop = field_offset + field_length - indicator_length + 1
            position = start

            # ищем значение поля до первого разделителя
            while position < stop:
                if record[start] == SUBFIELD_DELIMITER:
                    break
                position += 1

            # если есть текст до первого раздлителя, запоминаем его
            if position != start:
                field.value = record[start:position].decode(charset)

            # просматриваем подполя
            start = position
            while start < stop:
                position = start + 1
                while position < stop:
                    if record[position] == SUBFIELD_DELIMITER:
                        break
                    position += 1
                subfield = SubField(chr(record[start + 1]),
                                    record[start + 2:position].decode(charset))
                field.subfields.append(subfield)
                start = position

        # переходим к следующему полю в справочнике
        directory += 12

    return result


def write_iso_record(stream, record: Record, encoding: str) -> None:
    """
    Сохранение записи в файл в формате ISO 2709.

    :param stream: Поток
    :param record: Запись
    :param encoding: Кодировка
    :return: None
    """

    record_length = MARKER_LENGTH
    dictionary_length = 1  # С учетом ограничителя справочника
    field_length: 'List[int]' = []

    # Сначала подсчитываем общую длину записи
    for field in record.fields:

        if field.tag <= 0 or field.tag >= 1000:
            # Невозможно закодировать тег поля
            raise Exception

        dictionary_length += 12  # Одна статья справочника
        this_field_length = 0
        if field.tag < 10:
            # В фиксированном поле не бывает подполей и индикаторов
            val = field.value
            if val:
                this_field_length += len(val.encode(encoding))
        else:
            this_field_length += 2  # Индикаторы
            if field.value:
                this_field_length += len(field.value.encode(encoding))
            for subfield in field.subfields:
                code = subfield.code
                if code is None or ord(code) <= 32 or ord(code) >= 255:
                    raise IrbisError('Bad code: ' + safe_str(code))
                this_field_length += 2  # Признак подполя и его код
                val = subfield.value
                if val:
                    this_field_length += len(val.encode(encoding))
        this_field_length += 1  # Разделитель полей

        if this_field_length >= 10_000:
            # Слишком длинное поле
            raise Exception

        field_length.append(this_field_length)
        record_length += this_field_length

    record_length += dictionary_length  # Справочник
    record_length += 1  # Разделитель записей

    if record_length >= 100_000:
        # Слишком длинная запись
        raise Exception

    # Приступаем к кодированию
    dictionary_position = MARKER_LENGTH
    base_address = MARKER_LENGTH + dictionary_length
    current_address = base_address
    buffer = bytearray(record_length)
    for i in range(base_address):
        buffer[i] = 32  # Заполняем пробелами
    encode_int(buffer, 0, 5, record_length)
    encode_int(buffer, 12, 5, base_address)

    buffer[5] = ord('n')  # Record status
    buffer[6] = ord('a')  # Record type
    buffer[7] = ord('m')  # Bibliographical index
    buffer[8] = ord('2')
    buffer[10] = ord('2')
    buffer[11] = ord('2')
    buffer[17] = ord(' ')  # Bibliographical level
    buffer[18] = ord('i')  # Cataloging rules
    buffer[19] = ord(' ')  # Related record
    buffer[20] = ord('4')  # Field length
    buffer[21] = ord('5')  # Field offset
    buffer[22] = ord('0')

    # Кодируем конец справочника
    buffer[base_address - 1] = FIELD_DELIMITER

    # Проходим по полям
    for i, field in enumerate(record.fields):
        # Кодируем справочник
        encode_int(buffer, dictionary_position + 0, 3,
                   field.tag)
        encode_int(buffer, dictionary_position + 3, 4,
                   field_length[i])
        encode_int(buffer, dictionary_position + 7, 5,
                   current_address - base_address)

        # Кодируем поле
        if field.tag < 10:
            # В фиксированном поле не бывает подполей и индикаторов
            encode_str(buffer, current_address, field.value, encoding)
        else:
            # Два индикатора
            buffer[current_address + 0] = 32
            buffer[current_address + 1] = 32
            current_address += 2

            # Значение поля до первого разделителя
            current_address = encode_str(buffer, current_address,
                                         field.value, encoding)

            # Подполя
            for subfield in field.subfields:
                buffer[current_address + 0] = SUBFIELD_DELIMITER
                buffer[current_address + 1] = ord(subfield.code)
                current_address += 2
                current_address = encode_str(buffer, current_address,
                                             subfield.value, encoding)
        buffer[current_address] = FIELD_DELIMITER
        current_address += 1
        dictionary_position += 12

    # Ограничитель записи
    buffer[record_length - 2] = FIELD_DELIMITER
    buffer[record_length - 1] = RECORD_DELIMITER

    # Собственно записываем
    stream.write(buffer)


###############################################################################


__all__ = ['read_text_record', 'read_iso_record', 'STOP_MARKER',
           'write_text_record', 'write_iso_record']



# coding: utf-8

"""
Модуль реализующий объектно-ориентированное представление записи
с нераскодированными полями/подполями.
"""

from typing import TYPE_CHECKING
from irbis.records.abstract import AbstractRecord
if TYPE_CHECKING:
    from typing import List


class RawRecord(AbstractRecord):
    """
    Запись с нераскодированными полями/подполями.
    """
    __slots__ = 'database', 'mfn', 'status', 'version', 'fields'
    fields: 'List[str]'

    def __init__(self, *args: str) -> None:
        self.field_type = str
        super().__init__(*args)

    def __bulk_set__(self, *args: str):
        """
        Приватный метод установки полей записи.

        Внимание. Пользователь не должен явно обращаться к данному методу.

        :param args: список строк
        :return: ничего
        """
        if all((isinstance(arg, self.field_type) for arg in args)):
            self.fields += list(args)
        else:
            message = f'All args must be {self.field_type.__name__} type'
            raise TypeError(message)

    def clone_fields(self) -> 'List[str]':
        return self.fields.copy()

    def parse_line(self, line: str) -> None:
        self.fields.append(line)

    def __iter__(self):
        yield from self.fields



# coding: utf-8

"""
Статистика работы сервера.
"""

from typing import TYPE_CHECKING
from irbis.response import ServerResponse
if TYPE_CHECKING:
    from typing import Optional


class ClientInfo:
    """
    Информация о клиенте, подключенном к серверу ИРБИС
    (не обязательно о текущем)
    """

    __slots__ = ('number', 'ip_address', 'port', 'name', 'client_id',
                 'workstation', 'registered', 'acknowledged',
                 'last_command', 'command_number')

    def __init__(self):
        self.number: 'Optional[str]' = None
        self.ip_address: 'Optional[str]' = None
        self.port: 'Optional[str]' = None
        self.name: 'Optional[str]' = None
        self.client_id: 'Optional[str]' = None
        self.workstation: 'Optional[str]' = None
        self.registered: 'Optional[str]' = None
        self.acknowledged: 'Optional[str]' = None
        self.last_command: 'Optional[str]' = None
        self.command_number: 'Optional[str]' = None

    def __str__(self):
        return ' '.join([self.number, self.ip_address, self.port, self.name,
                         self.client_id, self.workstation,
                         self.registered, self.acknowledged,
                         self.last_command, self.command_number])


class ServerStat:
    """
    Статистика работы ИРБИС-сервера
    """

    __slots__ = 'running_clients', 'client_count', 'total_command_count'

    def __init__(self):
        self.running_clients: [ClientInfo] = []
        self.client_count: int = 0
        self.total_command_count: int = 0

    # noinspection DuplicatedCode
    def parse(self, response: ServerResponse) -> None:
        """
        Parse the server response for the stat.

        :param response: Server response
        :return: None
        """

        self.total_command_count = response.number()
        self.client_count = response.number()
        lines_per_client = response.number()
        if lines_per_client:
            for _ in range(self.client_count):
                client = ClientInfo()
                client.number = response.ansi()
                client.ip_address = response.ansi()
                client.port = response.ansi()
                client.name = response.ansi()
                client.client_id = response.ansi()
                client.workstation = response.ansi()
                client.registered = response.ansi()
                client.acknowledged = response.ansi()
                client.last_command = response.ansi()
                client.command_number = response.ansi()
                self.running_clients.append(client)
        else:
            # Ууточнить, требуется ли бросать исключение
            pass

    def __str__(self):
        return str(self.client_count) + ', ' + str(self.total_command_count)


__all__ = ['ClientInfo', 'ServerStat']
